using System;
using MediaPortal.GUI.Library;
using System.Runtime.InteropServices;

// Author: Roy Solberg <ultrax at roy solberg dot com>
namespace MediaPortal.InputDevices
{
    public class UltraXMediaRemotePlugin : ISetupForm, IPlugin, IPluginReceiver
    {
        protected const string LOG_PREFIX = "UltraXMediaRemotePlugin: ";
        private const int RIM_TYPEMOUSE = 0;
        private const int RIM_TYPEKEYBOARD = 1;
        private const int RIM_TYPEHID = 2;
        private const int RID_INPUT = 0x10000003;

        #region ISetupForm Members

        public string Author()
        {
            return "Roy Solberg - http://roysolberg.com/";
        }

        public bool CanEnable()
        {
            return true;
        }

        public bool DefaultEnabled()
        {
            return true;
        }

        public string Description()
        {
            return "Plugin for controlling MediaPortal using the Logitech UltraX Media Remote";
        }

        public bool GetHome(out string strButtonText, out string strButtonImage, out string strButtonImageFocus, out string strPictureImage)
        {
            strButtonText = String.Empty;
            strButtonImage = String.Empty;
            strButtonImageFocus = String.Empty;
            strPictureImage = String.Empty;
            return false;
        }

        public int GetWindowId()
        {
            return 4444;
        }

        public bool HasSetup()
        {
            return false;
        }

        public string PluginName()
        {
            return "UltraX Media Remote Plugin";
        }

        public void ShowPlugin()
        {
            return;
        }

        #endregion


        #region Structures

        public struct RAWINPUTDEVICE
        {
            public ushort usUsagePage;
            public ushort usUsage;
            public uint dwFlags;
            public IntPtr hwndTarget;
        }

        [StructLayout(LayoutKind.Explicit)]
        internal struct RAWINPUT
        {
            [FieldOffset(0)]
            public RAWINPUTHEADER header;
            [FieldOffset(16)]
            public RAWMOUSE mouse;
            [FieldOffset(16)]
            public RAWKEYBOARD keyboard;
            [FieldOffset(16)]
            public RAWHID hid;
        }


        [StructLayout(LayoutKind.Sequential)]
        internal struct RAWINPUTHEADER
        {
            [MarshalAs(UnmanagedType.U4)]
            public int dwType;
            [MarshalAs(UnmanagedType.U4)]
            public int dwSize;
            public IntPtr hDevice;
            [MarshalAs(UnmanagedType.U4)]
            public int wParam;
        }

        [StructLayout(LayoutKind.Sequential)]
        internal struct RAWKEYBOARD
        {
            [MarshalAs(UnmanagedType.U2)]
            public ushort MakeCode;
            [MarshalAs(UnmanagedType.U2)]
            public ushort Flags;
            [MarshalAs(UnmanagedType.U2)]
            public ushort Reserved;
            [MarshalAs(UnmanagedType.U2)]
            public ushort VKey;
            [MarshalAs(UnmanagedType.U4)]
            public uint Message;
            [MarshalAs(UnmanagedType.U4)]
            public uint ExtraInformation;
        }

        [StructLayout(LayoutKind.Explicit)]
        internal struct RAWMOUSE
        {
            [MarshalAs(UnmanagedType.U2)]
            [FieldOffset(0)]
            public ushort usFlags;
            [MarshalAs(UnmanagedType.U4)]
            [FieldOffset(4)]
            public uint ulButtons;
            [FieldOffset(4)]
            public BUTTONSSTR buttonsStr;
            [MarshalAs(UnmanagedType.U4)]
            [FieldOffset(8)]
            public uint ulRawButtons;
            [FieldOffset(12)]
            public int lLastX;
            [FieldOffset(16)]
            public int lLastY;
            [MarshalAs(UnmanagedType.U4)]
            [FieldOffset(20)]
            public uint ulExtraInformation;
        }

        [StructLayout(LayoutKind.Sequential)]
        internal struct RAWHID
        {
            [MarshalAs(UnmanagedType.U4)]
            public int dwSizHid;
            [MarshalAs(UnmanagedType.U4)]
            public int dwCount;
        }

        [StructLayout(LayoutKind.Sequential)]
        internal struct BUTTONSSTR
        {
            [MarshalAs(UnmanagedType.U2)]
            public ushort usButtonFlags;
            [MarshalAs(UnmanagedType.U2)]
            public ushort usButtonData;
        }

        #endregion

        #region DllImports
        [DllImport("User32.dll")]
        extern static bool RegisterRawInputDevices(RAWINPUTDEVICE[] pRawInputDevice, uint uiNumDevices, uint cbSize);

        [DllImport("User32.dll")]
        extern static uint GetRawInputData(IntPtr hRawInput, uint uiCommand, IntPtr pData, ref uint pcbSize, uint cbSizeHeader);

        #endregion DllImports

        #region Windows.h structure declarations

        // The following structures are defined in Windows.h

        [StructLayout(LayoutKind.Sequential)]
        internal struct RAWINPUTDEVICELIST
        {
            public IntPtr hDevice;
            [MarshalAs(UnmanagedType.U4)]
            public int dwType;
        }

        #endregion

        #region IPlugin Members

        public void Start()
        {
            Log.Debug(LOG_PREFIX + " Start()");

            RAWINPUTDEVICE[] rid = new RAWINPUTDEVICE[1];

            rid[0].usUsagePage = 0xFFBC; // Same as MS MCE remote
            rid[0].usUsage = 0x88; // Same as MS MCE remote
            rid[0].dwFlags = 0; // We only want to listen when MediaPortal is in focus
            rid[0].hwndTarget = GUIGraphicsContext.ActiveForm;
            if (!RegisterRawInputDevices(rid, (uint)rid.Length, (uint)Marshal.SizeOf(rid[0])))
            {
                throw new ApplicationException("Failed to register raw input device.");
            }
            else
            {
                Log.Debug(LOG_PREFIX + "UltraX Media Remote is registered and up and running.");
            }
        }


        public void Stop()
        {
            Log.Debug(LOG_PREFIX + " Stop()");
        }

        #endregion

        public bool WndProc(ref System.Windows.Forms.Message msg)
        {
            switch (msg.Msg)
            {
                case 0xFF: // Msg sent by the UltraX remote
                    uint dwSize = 0;
                    GetRawInputData(msg.LParam, RID_INPUT, IntPtr.Zero, ref dwSize, (uint)Marshal.SizeOf(typeof(RAWINPUTHEADER)));
                    IntPtr buffer = Marshal.AllocHGlobal((int)dwSize);
                    if (GetRawInputData(msg.LParam, RID_INPUT, buffer, ref dwSize, (uint)Marshal.SizeOf(typeof(RAWINPUTHEADER))) == dwSize)
                    {
                        RAWINPUT raw = (RAWINPUT)Marshal.PtrToStructure(buffer, typeof(RAWINPUT));
                        if (raw.header.dwType == RIM_TYPEHID)
                        {
                            // XXX: What about combinations? E.g. HOME + CLOSE = 0x2000104
                            switch (raw.keyboard.Message)
                            {
                                //
                                // Orange 'upper' buttons
                                //
                                case 0x1000004: // START
                                    GUIGraphicsContext.OnAction(new Action(Action.ActionType.ACTION_EXIT, 0, 0));
                                    return true;
                                case 0x104: // HOME
                                    GUIGraphicsContext.OnAction(new Action(Action.ActionType.ACTION_HOME, 0, 0));
                                    return true;
                                case 0x02000004: // CLOSE
                                    GUIGraphicsContext.OnAction(new Action(Action.ActionType.ACTION_CLOSE_DIALOG, 0, 0));
                                    return true;
                                case 0x204: // TV
                                    GUIWindowManager.ActivateWindow((int)GUIWindow.Window.WINDOW_TV);
                                    return true;
                                case 0x8004: // DVD
                                    GUIWindowManager.ActivateWindow((int)GUIWindow.Window.WINDOW_DVD);
                                    return true;
                                case 0x1004: // PICTURES
                                    GUIWindowManager.ActivateWindow((int)GUIWindow.Window.WINDOW_PICTURES);
                                    return true;
                                case 0x2004: // VIDEOS
                                    GUIWindowManager.ActivateWindow((int)GUIWindow.Window.WINDOW_VIDEOS);
                                    return true;
                                case 0x804: // MUSIC
                                    GUIWindowManager.ActivateWindow((int)GUIWindow.Window.WINDOW_MUSIC_FILES);
                                    return true;
                                //
                                // 'Lower' buttons
                                //
                                case 0x40004: // DVD MENU
                                    GUIGraphicsContext.OnAction(new Action(Action.ActionType.ACTION_CONTEXT_MENU, 0, 0));
                                    return true;
                                case 0x10004: // ANGLE
                                    return false;
                                case 0x20004: // LANGUAGE
                                    GUIGraphicsContext.OnAction(new Action(Action.ActionType.ACTION_AUDIO_NEXT_LANGUAGE, 0, 0));
                                    return true;
                                case 0x80004: // SUBTITLE
                                    GUIGraphicsContext.OnAction(new Action(Action.ActionType.ACTION_NEXT_SUBTITLE, 0, 0));
                                    return true;
                                case 0x100004: // SAP
                                    GUIGraphicsContext.OnAction(new Action(Action.ActionType.ACTION_ASPECT_RATIO, 0, 0));
                                    return true;
                                case 0x800004: // REPEAT
                                    return false;
                                case 0x404: // SHUFFLE
                                    return false;
                                case 0x400004: // LAST:CH
                                    GUIGraphicsContext.OnAction(new Action(Action.ActionType.ACTION_LAST_VIEWED_CHANNEL, 0, 0));
                                    return true;
                                case 0x200004: // CC/TELETEXT
                                    GUIWindowManager.ActivateWindow((int)GUIWindow.Window.WINDOW_TELETEXT);
                                    return true;
                                //
                                // Other
                                //
                                case 0x4: // Button released
                                    return false;
                                default: // Unknown
                                    Log.Info(LOG_PREFIX + "Uknown key [0x" + raw.keyboard.Message.ToString("X") + "]:[" + raw.keyboard.Message + "] pressed.");
                                    return false;

                            }
                        }
                        else
                        {
                            Log.Info(LOG_PREFIX + "Unhandled dwType:" + raw.header.dwType);
                        }
                    }

                    break;
                default:
                    break;
            }
            // XXX: What happens when INFO/EPG and FM RADIO is pressed?

            return false;
        }

    }
}
